#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <vector>
#include <map>
#include "pstream.h" // For reading gzip and bz2 compressed files

using namespace std;

string chainID_candidate="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                         "abcdefghijklmnopqrstuvexyz0123456789 ";

void print_help()
{
    cout <<
"Usage: cif2pdb PDBx.cif > PDBx.pdb\n"
"    Convert PDBx/mmCIF file PDBx.cif into PDB format file PDBx.pdb\n"
"\n"
"    Since it is not always possible to convert chain ID from PDBx/mmCIF\n"
"    to PDB, the chain ID in coverted PDB file is systematically renamed.\n"
    <<endl;
    exit(EXIT_SUCCESS);
}

/* split a long string into vectors by whitespace 
 * line          - input string
 * line_vec      - output vector 
 * delimiter     - delimiter */
void split(const string &line, vector<string> &line_vec,
    const char delimiter=' ')
{
    bool within_word = false;
    for (int pos=0;pos<line.size();pos++)
    {
        if (line[pos]==delimiter)
        {
            within_word = false;
            continue;
        }
        if (!within_word)
        {
            within_word = true;
            line_vec.push_back("");
        }
        line_vec.back()+=line[pos];
    }
}

void cif2pdb(const string filename,string &PDBtxt)
{
    bool loop_ = false; // not reading following content
    map<string,int> _atom_site;
    map<string,char> asym_id2chainID;
    char chainID=' ';
    int atom_site_pos;
    vector<string> line_vec;
    string line;
    int a=0; // atom index
    string alt_id=".";  // alternative location indicator
    string asym_id="."; // this is similar to chainID, except that
                        // chainID is char while asym_id is a string
                        // with possibly multiple char
    string AA="";       // residue name
    string atom="";
    string resi="";
    string model_index=""; // the same as model_idx but type is string
    string group_PDB="";   // ATOM/HETATM
    float Cartn_x, Cartn_y, Cartn_z;
    float B_iso_or_equiv=0;
    string type_symbol=""; // element symbol
    stringstream i8_stream;


    int filetype=0; // 0 - uncompressed; 1 - gz; 2 - bz2; -1 - stdin
    ifstream fin;
    redi::ipstream fin_gz; // if file is compressed
    if (filename.size()>=3 && 
        filename.substr(filename.size()-3,3)==".gz")
    {
        fin_gz.open("zcat "+filename);
        filetype=1;
    }
    else if (filename.size()>=4 && 
        filename.substr(filename.size()-4,4)==".bz2")
    {
        fin_gz.open("bzcat "+filename);
        filetype=2;
    }
    else if (filename=="-") filetype=-1;
    else fin.open(filename.c_str());


    while (filetype==-1?cin.good():(filetype?fin_gz.good():fin.good()))
    {
        if (filetype==-1)  getline(cin, line);
        else if (filetype) getline(fin_gz, line);
        else                    getline(fin, line);
        if (loop_) loop_ = line.compare(0,2,"# ");
        if (!loop_)
        {
            if (line.compare(0,5,"loop_")) continue;
            if (filetype==-1)  getline(cin, line);
            else if (filetype) getline(fin_gz, line);
            else               getline(fin, line);
            if (line.compare(0,11,"_atom_site.")) continue;

            loop_=true;
            _atom_site.clear();
            atom_site_pos=0;
            _atom_site[line.substr(11,line.size()-12)]=atom_site_pos;

            while(1)
            {
                if (filetype==-1)  getline(cin, line);
                else if (filetype) getline(fin_gz, line);
                else               getline(fin, line);
                if (line.compare(0,11,"_atom_site.")) break;
                _atom_site[line.substr(11,line.size()-12)]=++atom_site_pos;
            }


            if (_atom_site.count("group_PDB")*
                _atom_site.count("label_atom_id")*
                _atom_site.count("label_comp_id")*
               (_atom_site.count("auth_asym_id")+
                _atom_site.count("label_asym_id"))*
               (_atom_site.count("auth_seq_id")+
                _atom_site.count("label_seq_id"))*
                _atom_site.count("Cartn_x")*
                _atom_site.count("Cartn_y")*
                _atom_site.count("Cartn_z")==0)
            {
                loop_ = false;
                cerr<<"Warning! Missing one of the following _atom_site data items: group_PDB, label_atom_id, label_atom_id, auth_asym_id/label_asym_id, auth_seq_id/label_seq_id, Cartn_x, Cartn_y, Cartn_z"<<endl;
                continue;
            }
        }

        line_vec.clear();
        split(line,line_vec);
        group_PDB=line_vec[_atom_site["group_PDB"]];
        if (group_PDB!="ATOM" && group_PDB!="HETATM") continue;
        if (group_PDB=="ATOM") group_PDB="ATOM  ";
        
        /* remove alternatively located residue */
        alt_id=".";
        if (_atom_site.count("label_alt_id")) // in 39.4 % of entries
            alt_id=line_vec[_atom_site["label_alt_id"]];
        if (alt_id!="." && alt_id!="A") continue;

        atom=line_vec[_atom_site["label_atom_id"]];
        if (atom[0]=='"') atom=atom.substr(1);
        if (atom.size() && atom[atom.size()-1]=='"')
            atom=atom.substr(0,atom.size()-1);
        if (atom.size()==0) continue;
        if      (atom.size()==1) atom=" "+atom+"  ";
        else if (atom.size()==2) atom=" "+atom+" "; // wrong for sidechain H
        else if (atom.size()==3) atom=" "+atom;
        else if (atom.size()>=5) atom=atom.substr(0,4);

        AA=line_vec[_atom_site["label_comp_id"]]; // residue name
        if      (AA.size()==1) AA="  "+AA;
        else if (AA.size()==2) AA=" " +AA;
        else if (AA.size()>=4) AA=AA.substr(0,3);

        if (_atom_site.count("pdbx_PDB_model_num") && 
            model_index!=line_vec[_atom_site["pdbx_PDB_model_num"]])
        {
            if (model_index!="") PDBtxt+="ENDMDL\n";
            model_index=line_vec[_atom_site["pdbx_PDB_model_num"]];
            i8_stream<<"MODEL "<<setw(4)<<model_index<<'\n';
            PDBtxt+=i8_stream.str();
            i8_stream.str(string());

            a=0;
            asym_id2chainID.clear();
        }

        if (_atom_site.count("auth_asym_id"))
             asym_id=line_vec[_atom_site["auth_asym_id"]];
        else asym_id=line_vec[_atom_site["label_asym_id"]];
        if (asym_id2chainID.count(asym_id)==0)
        {
            chainID=chainID_candidate[
                asym_id2chainID.size() % chainID_candidate.size()];
            asym_id2chainID[asym_id]=chainID;
        }

        if (_atom_site.count("auth_seq_id"))
             resi=line_vec[_atom_site["auth_seq_id"]];
        else resi=line_vec[_atom_site["label_seq_id"]];
        if (_atom_site.count("pdbx_PDB_ins_code") && 
            line_vec[_atom_site["pdbx_PDB_ins_code"]]!="?")
            resi+=line_vec[_atom_site["pdbx_PDB_ins_code"]][0];
        else resi+=" ";

        Cartn_x=atof(line_vec[_atom_site["Cartn_x"]].c_str());
        Cartn_y=atof(line_vec[_atom_site["Cartn_y"]].c_str());
        Cartn_z=atof(line_vec[_atom_site["Cartn_z"]].c_str());

        B_iso_or_equiv=0;
        if (_atom_site.count("B_iso_or_equiv")) B_iso_or_equiv=
            atof(line_vec[_atom_site["B_iso_or_equiv"]].c_str());
        
        a++;
        i8_stream<<group_PDB
            <<setw(5)<<a % 100000<<" "<<atom<<" "<<AA<<" "
            <<asym_id2chainID[asym_id]<<setw(5)<<resi<<"   "
            <<setiosflags(ios::fixed)<<setprecision(3)
            <<setw(8)<<Cartn_x<<setw(8)<<Cartn_y<<setw(8)<<Cartn_z
            <<"  1.00"<<setiosflags(ios::fixed)<<setprecision(2)
            <<setw(6)<<B_iso_or_equiv;
        if (_atom_site.count("type_symbol"))
            i8_stream<<setw(12)<<line_vec[_atom_site["type_symbol"]];
        PDBtxt+=i8_stream.str()+"\n";
        i8_stream.str(string());
    }
    if (model_index!="") PDBtxt+="ENDMDL\n";
    _atom_site.clear();
    line_vec.clear();
    line.clear();

    if (filetype>0) fin_gz.close();
    else            fin.close();
}

int main(int argc, char *argv[])
{
    /* get argument */
    if (argc < 2) print_help();
    string filename=argv[1];
    if (filename=="-h") print_help();

    /* convert format */
    string PDBtxt;
    cif2pdb(filename,PDBtxt);
    cout<<PDBtxt<<"END"<<endl;
    PDBtxt.clear();
    return 0;
}
